
class BytesAndInts:

    def __init__(self):
        pass

    @staticmethod
    def int2Byte(i: int) -> bytes:
        return i.to_bytes(i.bit_length(), "little").rstrip(b'\x00')

    @staticmethod
    def byte2Int(b: bytes) -> int:
        return int.from_bytes(b, "little")



from __future__ import annotations
import json
import secrets
from AsymmetricEncryptions.General import PrimeNumberGen, XOR
import hmac
import hashlib

class ElGamalKey:

    def __init__(self, p: int, g: int, y: int, x: int or None = None) -> None:
        self.p: int = p
        self.g: int = g
        self.y: int = y
        self.x: int = x
        self.has_private: bool = False
        if x:
            self.has_private = True
            self.public: ElGamalKey = ElGamalKey(p=p, g=g, y=y, x=None)

    def __eq__(self, other):
        if not isinstance(other, ElGamalKey): return False
        return hashlib.sha256(f"{self}".encode()).hexdigest() == hashlib.sha256(f"{other}".encode()).hexdigest()

    def __str__(self) -> str:
        r: str = ""
        if self.has_private:
            r += f"""
Private Key:

p = {self.p}
q = {self.g}
x = {self.x}
y = {self.y}
"""
        r += f"""
Public Key:

p = {self.p}
q = {self.g}
y = {self.y}
"""
        return r

    @staticmethod
    def new(nBit) -> ElGamalKey:
        p: int = PrimeNumberGen.generate(nBit)
        g: int = secrets.randbelow(p)
        x: int = secrets.randbelow(p)
        y: int = pow(g, x, p)
        return ElGamalKey(p=p, g=g, y=y, x=x)


    def export(self, file_name: str, pwd: bytes = b"\x00") -> None:
        data_dict: dict = {"p": self.p, "g": self.g, "y": self.y, "x": self.x}
        jData: bytes = json.dumps(data_dict).encode("utf-8")
        write_data: bytes = XOR.repeated_key_xor(jData, pwd)
        mac: hmac = hmac.new(key=pwd, msg=jData, digestmod="sha512")
        with open(file_name, "wb") as f:
            f.write(mac.digest() + write_data)

    @staticmethod
    def load(file_name: str, pwd: bytes = b"\x00") -> ElGamalKey:
        with open(file_name, "rb") as f:
            dMac: bytes = f.read(64)
            read_data: bytes = f.read()
        jData: bytes = XOR.repeated_key_xor(read_data, pwd)
        mac: hmac = hmac.new(key=pwd, msg=jData, digestmod="sha512")
        assert mac.digest() == dMac
        return ElGamalKey(**json.loads(jData.decode()))



from __future__ import annotations
from .EllipticCurveNISTP256 import EllipticCurveNISTP256
import secrets
class ECKey:
    def __init__(self) -> None:
        self.private_key = secrets.randbelow(EllipticCurveNISTP256().n)
        self.public_key = EllipticCurveNISTP256().g() * self.private_key

    def __str__(self):
        return f"{self.private_key = }\nself.public_key = {self.public_key }"



from __future__ import annotations
from AsymmetricEncryption.AsymmetricEncryptions.General import PrimeNumberGen
import secrets

class DLIESKey:

    def __init__(self, g: int, n: int, y: int, x: int = None):
        self.g: int = g
        self.n: int = n
        self.y: int = y
        self.x: int = x
        self.has_private: bool = False
        if x:
            self.has_private = True
            self.public: DLIESKey = DLIESKey(g, n, y)

    @staticmethod
    def new(nBit: int) -> DLIESKey:
        n: int = PrimeNumberGen.generate(nBit)
        g: int = secrets.randbelow(n)
        x: int = secrets.randbelow(n)
        y: int = pow(g, x, n)
        return DLIESKey(g, n, y, x)



from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.RSA.RSAKey import RSAKey
from AsymmetricEncryptions.General import BytesAndInts

class RSA:
    def __init__(self, key: RSAKey) -> None:
        self.key = key

    @staticmethod
    def generate_key_pair(nBit) -> RSAKey and RSAKey:
        # RSAKey.new returns a pair of priv and pub
        Priv: RSAKey = RSAKey.new(nBit)
        Pub: RSAKey = Priv.public
        return Priv, Pub


    def encrypt(self, msg: bytes, assertion=True) -> bytes:
        int_msg: int = BytesAndInts.byte2Int(msg)
        if assertion: assert int_msg < self.key.n
        cipher: int = pow(int_msg, self.key.e, self.key.n)
        return BytesAndInts.int2Byte(cipher)

    def decrypt(self, cipher: bytes) -> bytes:
        assert self.key.has_private
        int_cipher: int = BytesAndInts.byte2Int(cipher)
        m: int = pow(int_cipher, self.key.d, self.key.n)
        return BytesAndInts.int2Byte(m)

    def sign(self, msg: bytes, assertion=True) -> bytes:
        assert self.key.has_private
        int_msg: int = BytesAndInts.byte2Int(msg)
        if assertion: assert int_msg < self.key.n
        s: int = pow(int_msg, self.key.d, self.key.n)
        return BytesAndInts.int2Byte(s)

    def verify(self, s, og_msg: bytes) -> None:
        int_s: int = BytesAndInts.byte2Int(s)
        cipher: int = pow(int_s, self.key.e, self.key.n)
        assert cipher == BytesAndInts.byte2Int(og_msg
                                               )


from AsymmetricEncryptions.General import XOR, BytesAndInts
from .DLIESKey import DLIESKey
import secrets

class DLIES:

    def __init__(self) -> None:
        pass

    @staticmethod
    def encrypt(pub_key: DLIESKey, msg: bytes, encryption_func=XOR.repeated_key_xor) -> tuple[bytes, int]:
        r: int = secrets.randbelow(pub_key.n)
        R: int = pow(pub_key.g, r, pub_key.n)
        S: int = pow(pub_key.y, r, pub_key.n)
        S_bytes: bytes = BytesAndInts.int2Byte(S)
        encrypted = encryption_func(msg, S_bytes)
        return encrypted, R

    @staticmethod
    def decrypt(private_key: DLIESKey, ciphertxt: tuple[bytes, int], decryption_function=XOR.repeated_key_xor) -> bytes:
        encrypted, R = ciphertxt
        assert private_key.x and private_key.has_private
        S: int = pow(R, private_key.x, private_key.n)
        S_bytes: bytes = BytesAndInts.int2Byte(S)
        return decryption_function(encrypted, S_bytes)
